/* packet-multipart_mixed.c
 * Routines for multipart_mixed_media_type dissection
 * Copyright 2004, Anders Broman <anders.broman[at]ericsson.com>
 *
 * $Id: packet-multipart_mixed,v 1.0 2004/01/09 23:33:00 guy Exp $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 *
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * References for "media-type multipart/mixed :
 * http://www.iana.org/assignments/media-types/index.html
 * http://www.rfc-editor.org/rfc/rfc2045.txt
 * http://www.rfc-editor.org/rfc/rfc2046.txt
 * http://www.rfc-editor.org/rfc/rfc2047.txt
 * http://www.rfc-editor.org/rfc/rfc2048.txt
 * http://www.rfc-editor.org/rfc/rfc2049.txt
 *
 * Part of the code is modeled from the SIP dissector
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "prefs.h"
#include <glib.h>

#ifdef NEED_SNPRINTF_H
# include "snprintf.h"
#endif

#include <epan/packet.h>

/* Initialize the protocol and registered fields */
static int proto_multipart_mixed = -1;

/* Initialize the subtree pointers */
static gint ett_multipart_mixed = -1;

/* Not shure that compact_name exists for multipart, but choose to keep 
 * the structure from SIP dissector, all the content- is also from SIP */

typedef struct {
        char *name;
        char *compact_name;
} multipart_header_t;

static const multipart_header_t multipart_headers[] = {
                { "Unknown-header", 		NULL }, /* Pad so that the real headers start at index 1 */
                { "Content-Disposition", 	NULL },
                { "Content-Encoding", 		"e"  },
                { "Content-Language", 		NULL },
                { "Content-Length", 		"l"  },
                { "Content-Type", 			"c"  },
};

#define POS_CONTENT_DISPOSITION		1
#define POS_CONTENT_ENCODING		2
#define POS_CONTENT_LANGUAGE		3
#define POS_CONTENT_LENGTH			4
#define POS_CONTENT_TYPE			5

static gint hf_header_array[] = {
				-1, /* "Unknown-header" - Pad so that the real headers start at index 1 */
                -1, /* "Content-Disposition" */
                -1, /* "Content-Encoding" */
                -1, /* "Content-Language" */
                -1, /* "Content-Length" */
                -1, /* "Content-Type" */
};

/* Define media_type/Content type table */
static dissector_table_t media_type_dissector_table;

/* Determins if	bodies with no media type dissector shoud be displayed
 * as raw text, may cause problems with images sound etc
 * TODO improve to check for different content types ?
 */

static gboolean display_unknown_body_as_text = FALSE;


static gint find_first_boundary(tvbuff_t *tvb,char *buf, gint boundary_id_length, gint offset);
static gint find_next_boundary(tvbuff_t *tvb,char *buf, gint boundary_id_length, gint offset);
static void dissect_header_and_body(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, gint offset,gint body_end_offset);
static gint is_known_multipart_header(tvbuff_t *tvb, int offset,guint header_len);

/* Code to actually dissect the packets */

static void
dissect_multipart_mixed(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{

	char *content_type_par_str = NULL;
	char *boundary_id_start;
	char *boundary_id_end;
	gint boundary_id_length;
	gint body_end_offset;
	gint header_start_offset;

/* Boundary delimiters must not appear within the encapsulated material,
*  and must be no longer than 70 characters, not counting the two
*  leading hyphens.(quote from fc2046) */

	guchar buf[70];
	gint offset = 0;
	gint linelen;
	gint next_offset = 0;
	gint boundary_str_end;
	gboolean last_boundary = FALSE;
	gboolean found_match = FALSE;

	content_type_par_str = pinfo->private_data;
	if ( strncmp(content_type_par_str,"boundary=",9) !=0){
				proto_tree_add_text(tree,
		    tvb, offset, -1,"failed to find the boundary string in: %s",content_type_par_str);  

		return;
	}

 	if ( strchr(content_type_par_str,'"') !=NULL) {
		boundary_id_start = strchr(content_type_par_str,'"') +1;
 		boundary_id_end = strchr(boundary_id_start,'"');
	}
	else{
		boundary_id_start = strchr(content_type_par_str,'=') +1;
		/* not shure about this ?, how does string without " end ? */
		boundary_id_end = strchr(boundary_id_start,'/t');
	}
	boundary_id_length = boundary_id_end - boundary_id_start;
	memcpy(buf,boundary_id_start,boundary_id_length);
	buf[boundary_id_length] = '\0';

	proto_tree_add_text(tree, tvb, offset, -1,
			"Multipart/mixed body:");

/* Make no entries in Protocol column and Info column on summary display,but stop sub-dissectors
 *	from clearing entered text	*/

	if (check_col(pinfo->cinfo, COL_PROTOCOL)) 
		   col_set_fence(pinfo->cinfo, COL_INFO);

	offset = 0;
    while (tvb_offset_exists(tvb, offset)) {
		header_start_offset = find_first_boundary(tvb, buf, boundary_id_length, offset);
		linelen = header_start_offset - offset;

		/* Print boundary string + characters outside of body/header */
 		proto_tree_add_text(tree, tvb, offset, linelen,
			"%s", tvb_format_text(tvb, offset, linelen));

		proto_tree_add_text(tree, tvb, offset, -1,"Header:");

		body_end_offset = find_next_boundary(tvb, buf, boundary_id_length, header_start_offset);
		offset = body_end_offset;
		boundary_str_end = body_end_offset + 4 + boundary_id_length;
		if ( tvb_strneql(tvb, boundary_str_end, "--", 2) == 0) {
			/* end of bodies */
			offset = tvb_reported_length(tvb);
			last_boundary = TRUE;
		}	

		/* Dissect content header and body here */ 
		dissect_header_and_body(tvb, pinfo, tree, header_start_offset, body_end_offset);

		/* continue with next part*/
		if ( last_boundary ) {
			/* Print boundary string and the rest of tvb */
			linelen = offset - body_end_offset;
	 		proto_tree_add_text(tree, tvb, body_end_offset, linelen,
				"%s", tvb_format_text(tvb, body_end_offset, linelen));
		}
		
	}



/* If this protocol has a sub-dissector call it here, see section 1.8 */
}
static gint
find_first_boundary(tvbuff_t *tvb,char *buf, gint boundary_id_length, gint offset)
{
	gint next_offset = 0;
	gint linelen;
		
	    while (tvb_offset_exists(tvb, offset)) {
		tvb_find_line_end(tvb, offset, -1, &next_offset, FALSE);
	        linelen = next_offset - offset;
		    if ( tvb_strneql(tvb, offset, "--", 2) == 0) {
				if (tvb_strneql(tvb, offset + 2, buf, boundary_id_length) == 0) {
 					/* Next content type starts here */
					return next_offset;
				}
			}
			offset = next_offset;
		}
		return -1;
}
static gint
find_next_boundary(tvbuff_t *tvb,char *buf, gint boundary_id_length, gint offset)
{
	gint next_offset = 0;
	gint linelen;
		
	    while (tvb_offset_exists(tvb, offset)) {
		tvb_find_line_end(tvb, offset, -1, &next_offset, FALSE);
	        linelen = next_offset - offset;
		    if ( tvb_strneql(tvb, offset, "--", 2) == 0) {
				if (tvb_strneql(tvb, offset + 2, buf, boundary_id_length) == 0) {
 					/* Next content type starts here */
					return offset - 2;/* previous two char must be CRLF */
				}
			}
			offset = next_offset;
		}
		return -1;
}

static void
dissect_header_and_body(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, gint offset,gint body_end_offset)
{
	gint linelen;
	gint next_offset;
	gint body_length;
	char *media_type_str = NULL;
	char *content_type_par_str = NULL;
    tvbuff_t *next_tvb;
	proto_item *ti;
	proto_tree *multipart_mixed_tree;
	gboolean found_match;

	/* Code from packet-sip */
	while ( body_end_offset > offset) {
	    gint line_end_offset;
		gint colon_offset;
		gint semicolon_offset;
		gint content_type_len, content_type_par_str_len;
		gint header_len;
		gint hf_index;
		gint value_offset;
		guchar c;
		size_t value_len;
		char *value;

        linelen = tvb_find_line_end(tvb, offset, -1, &next_offset,
                  FALSE);
        if (linelen == 0) {
			/*
             * This is a blank line separating the
             * message header from the message body.
             */
			offset = next_offset; /* offset points to body start */
			break;
             
 		}
        line_end_offset = offset + linelen;
		colon_offset = tvb_find_guint8(tvb, offset, linelen, ':');
		if (colon_offset == -1) {
			/*
			 * Malformed header - no colon after the name.
			 */
			proto_tree_add_text(tree, tvb, offset,
			    next_offset - offset, "%s",
			    tvb_format_text(tvb, offset, linelen));
		} else {
			header_len = colon_offset - offset;

			hf_index = is_known_multipart_header(tvb,
			    offset, header_len);
			
			if (hf_index == -1) {
				proto_tree_add_text(tree, tvb,
				    offset, next_offset - offset, "%s",
				    tvb_format_text(tvb, offset, linelen));
			} else {
				/*
				 * Skip whitespace after the colon.
				 */
				value_offset = colon_offset + 1;
				while (value_offset < line_end_offset
				    && ((c = tvb_get_guint8(tvb,
					    value_offset)) == ' '
				      || c == '\t'))
					value_offset++;
				/*
				 * Fetch the value.
				 */
				value_len = line_end_offset - value_offset;
				value = tvb_get_string(tvb, value_offset,
				    value_len);

				/*
				 * Add it to the protocol tree,
				 * but display the line as is.
				 */
				
				switch ( hf_index ) {
				case POS_CONTENT_TYPE :
					proto_tree_add_string_format(tree,
					    hf_header_array[hf_index], tvb,
					    offset, next_offset - offset,
					    value, "%s",
					    tvb_format_text(tvb, offset, linelen));
					content_type_len = value_len;
					semicolon_offset = tvb_find_guint8(tvb, value_offset,linelen, ';');
					if ( semicolon_offset != -1) {
						content_type_len = semicolon_offset - value_offset;

						content_type_par_str_len = line_end_offset - ( semicolon_offset + 1);
						content_type_par_str = tvb_get_string(tvb, semicolon_offset + 1,
							content_type_par_str_len );
 					}
					media_type_str = tvb_get_string(tvb, value_offset, content_type_len);
					
					break;
				default :
					proto_tree_add_string_format(tree,
					    hf_header_array[hf_index], tvb,
					    offset, next_offset - offset,
					    value, "%s",
					    tvb_format_text(tvb, offset, linelen));
				break;
				}/* end switch */
				g_free(value);
			}/*if HF_index */

		}/* if colon_offset */ 
 			offset = next_offset;
        }/* End while */
		body_length = body_end_offset - offset;

		next_tvb = tvb_new_subset(tvb, offset, body_length, body_length);


		/* create display subtree for the protocol  		*/
		ti = proto_tree_add_text(tree, next_tvb, 0, -1,
                                         "Body:");
		multipart_mixed_tree = proto_item_add_subtree(ti, ett_multipart_mixed);

		/* give the content type parameters to sub dissectors */
		
		if ( media_type_str != NULL ) {
			pinfo->private_data = content_type_par_str; 
			found_match = dissector_try_string(media_type_dissector_table,
							   media_type_str,
							   next_tvb, pinfo,
							   multipart_mixed_tree);
			g_free(media_type_str);
			g_free(content_type_par_str);
			/* If no match dump as text */
		}
		if ( found_match != TRUE )
			/* There is no dissector for this kind of content dump it as text ? */
		{
			if ( display_unknown_body_as_text ) {
				offset = 0;
	        		while (tvb_offset_exists(next_tvb, offset)) {
	                	tvb_find_line_end(next_tvb, offset, -1, &next_offset, FALSE);
	                	linelen = next_offset - offset;
	                	proto_tree_add_text(multipart_mixed_tree, next_tvb, offset, linelen,
	                   		"%s", tvb_format_text(next_tvb, offset, linelen));
 	        			offset = next_offset;
 	 				}/* end while */
			}
			else {
				offset = 0;
				proto_tree_add_text(multipart_mixed_tree, next_tvb, offset, -1,
					"Body with no media type dissector:");
			}

		}
        

}

/* Returns index of method in multipart_headers */

static gint is_known_multipart_header(tvbuff_t *tvb, int offset, guint header_len)
{
        guint i;

        for (i = 1; i < array_length(multipart_headers); i++) {
                if (header_len == strlen(multipart_headers[i].name) &&
                    tvb_strncaseeql(tvb, offset, multipart_headers[i].name, header_len) == 0)
                        return i;
                if (multipart_headers[i].compact_name != NULL &&
                    header_len == strlen(multipart_headers[i].compact_name) &&
                    tvb_strncaseeql(tvb, offset, multipart_headers[i].compact_name, header_len) == 0)
                        return i;
 		}

 		return -1;
 }

/* Register the protocol with Ethereal */

/* this format is require because a script is used to build the C function
   that calls all the protocol registration.
*/

void
proto_register_multipart_mixed(void)
{                 

/* Setup list of header fields  See Section 1.6.1 for details*/
	static hf_register_info hf[] = {
                { &hf_header_array[POS_CONTENT_DISPOSITION],
		       { "Content-Disposition", "multipart_mixed.Content-Disposition", 
		       FT_STRING, BASE_NONE,NULL,0x0,
			"RFC 3261: Content-Disposition Header", HFILL }
		},
                { &hf_header_array[POS_CONTENT_ENCODING],
		       { "Content-Encoding","multipart_mixed.Content-Encoding", 
		       FT_STRING, BASE_NONE,NULL,0x0,
			"RFC 3261: Content-Encoding Header", HFILL }
		},
                { &hf_header_array[POS_CONTENT_LANGUAGE],
		       { "Content-Language","multipart_mixed.Content-Language", 
		       FT_STRING, BASE_NONE,NULL,0x0,
			"RFC 3261: Content-Language Header", HFILL }
		},
                { &hf_header_array[POS_CONTENT_LENGTH],
		       { "Content-Length", "multipart_mixed.Content-Length", 
		       FT_STRING, BASE_NONE,NULL,0x0,
			"RFC 3261: Content-Length Header", HFILL }
		},
                { &hf_header_array[POS_CONTENT_TYPE],
		       { "Content-Type", "multipart_mixed.Content-Type", 
		       FT_STRING, BASE_NONE,NULL,0x0,
			"RFC 3261: Content-Type Header", HFILL }
 		},
 	};
	
	module_t *multipart_mixed_module;
/* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_multipart_mixed,
	};

/* Register the protocol name and description */
	proto_multipart_mixed = proto_register_protocol("multipart/mixed media type",
	    "multipart_mixed", "multipart_mixed");

/* Required function calls to register the header fields and subtrees used 	*/
	proto_register_field_array(proto_multipart_mixed, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));

/* Get the content type and internet media type table */
	media_type_dissector_table = find_dissector_table("media_type");

	multipart_mixed_module = prefs_register_protocol(proto_multipart_mixed, NULL);

		prefs_register_bool_preference(multipart_mixed_module, "display_unknown_body_as_text",
		"Display bodies with no media type dissector as raw text",
		"may cause problems with images sound etc",
		&display_unknown_body_as_text);
}


/* If this dissector uses sub-dissector registration add a registration routine.
   This format is required because a script is used to find these routines and
   create the code that calls these routines.
*/
void
proto_reg_handoff_multipart_mixed(void)
{
	dissector_handle_t multipart_mixed_handle;

	multipart_mixed_handle = create_dissector_handle(dissect_multipart_mixed,
	    proto_multipart_mixed);
	dissector_add_string("media_type",
		"multipart/mixed", multipart_mixed_handle);
}
